/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 

/*
 Author: Michael Rao
*/

/* changelog: 
   V0.3: 21/07/11
   - interactive mode
   - use less memory

   V0.2: 15/07/11
   - new algo for sim_dir (quicker) 
   - use less memory

   V0.1: 10/07/11
   - initial version
*/

/*
  todo:
  - keep hash in separate file

  - sim dirs...
  - print % in sim_dir
  - discard useless sim_dir ? sort by % ?
  
  - virer O_DIRECT ?

  - mem: 
  -- ptr sur int
  -- list<tree_t> -> tab 
*/

#include <stdio.h>
#include <stdlib.h>
#include <map> 
#include <vector> 
#include <list> 
#include <set> 
#include <string> 
#include <string.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <openssl/md5.h>
#include <dirent.h> 
#include <fcntl.h>
#include <assert.h>

using namespace std;

void printusage() {
  printf("fdupdirs V 0.2\n" 
	 "Searches and reports duplicate files and directories (-H for more infos)\n"
	 "Usage: fdupdirs [options] path [ path [...] ]\n"
	 "\n"
	 "Options:\n"
	 " -h : print this help\n"
	 " -H : print more infos on this program\n"
	 " -i : interactive mode (use with care)\n"
	 " -n : do not perform a md5sum on whole file\n"
	 "      (use only file size, first, mid and last blocks)\n"
	 " -f x : cmpdir factor at least x (between 0 and 1, default: 0.05)\n"
	 "        close to 0 will find more sim dirs, but take more time.\n"
	 " -F x : report cmpdir factor at least x (between 0 and 1, default: 0.5)\n"
	 " -t : sort dirs by size (instead of saved space)\n"
	 " -r x : report duplicates of size at least x bytes (default: 2^20)\n"
	 " -s x : perform an prehash for every file of size at least x bytes (default: 2^20)\n"
	 "advanced :\n" 
	 " -b : use only the first 4k block to perform prehash\n"
	 " --onlysize : use only file size\n"
	 " --fakehash : assume that every file is different (testing purpose)\n"
	 " --fake : do not modify files (testing purpose)\n"
	 );
}

void printreadme() {
  printf(
	 "Searches and reports duplicate files and directories:\n"
	 "- two files are identicals if the MD5 hash is the same (in id_files.txt)\n"
	 "- reports two dirs as identical if they have the same set of hash files (ignore filenames, empty files and dirs) (in id_dirs.txt)\n"
	 "- if the two diretories are realy equal (same filenames, empty files ans dirs), it reports it with \"EXACT\" flag\n"
	 "- try to find \"similar\" dirs: dir A is a subset of dir B, or A inter B is big\n"
	 "- reports are sorted by the saved space size (reverse order)\n"
	 "\n"
	 "- do not consider or follow symlinks, or block/char files\n"
	 "- do not report hardlinked files as duplicate\n"
	 "- do not make a byte to byte comparaison (useless...)\n"
	 "\n"
	 "It tries to optimise the hash computation :\n" 
	 "- on a first pass, it compute only hash of first and last 4k data of larges files\n"
	 "- sort files by inode number before hashing\n"
	 "\n"
	 "Also:\n" 
	 "- report hard links in hl.txt\n"
	 "- report directories with a lot of files in lof_dirs.txt\n"
	 "- reports different files with same begining in sim_files.txt (useful for mp3 and containers with tags)\n"   
	 "\n"
	 "Known bugs :\n"
	 "- may use a lot of memory...\n"
	 "- search fo similar dirs may take time...\n"
	 "- EXACT dirs may be badly reported if there is more than two identical dirs\n"
	 );
}

typedef long long ll_t;

#define REAL 1
bool fakemod=0;
bool forcedel=0;

FILE *mod_log=NULL;
FILE *out_log=NULL;

#define lprintf(...) {printf ( __VA_ARGS__); if(out_log) fprintf(out_log,__VA_ARGS__);}

#define eprintf(...) {fprintf (stderr, __VA_ARGS__); if(out_log) fprintf(out_log,__VA_ARGS__);}


void *mbrk;
void printmemusage() {
  lprintf("mem: %Ld MB\n",ll_t((char*)sbrk(0)-(char*)mbrk)/1024/1024);
}

char *human(ll_t x,char *bf=NULL,int s=30) 
{
  static char bfhuman[30];
  if(bf==NULL) bf=bfhuman;
  if(x<1000) snprintf(bf,s,"%Ldb",x);
  else if(x<1000LL*1024) snprintf(bf,s,"%.1fk",x/1024.);
  else if(x<1000LL*1024*1204) snprintf(bf,s,"%.1fM",x/1024./1024.);
  else if(x<1000LL*1024*1204*1024) snprintf(bf,s,"%.1fG",x/1024./1024./1024.);
  else snprintf(bf,s,"%.1fT",x/1024./1024./1024./1024.);
  return bf;
}

set<char *> mstrdup_bf;

void mstrdup_free()
{
  set<char *>::iterator it;
  for(it= mstrdup_bf.begin();it!= mstrdup_bf.end();it++)
    delete [] *it;
  mstrdup_bf.clear();
}

char *mstrdup(const char *f)
{
  int s=1+strlen(f);
  static char *b=NULL;
  static int bs;
  if(b==NULL || bs<s) {
    bs=1024*1024;
    b=new char[bs];
    mstrdup_bf.insert(b);
  }
  char *r=b;
  memcpy(b,f,s);
  b+=s;
  bs-=s;
  return r;
}

bool md5=true;
bool noend=false;
ll_t systematic=1024LL*1024;
ll_t minreportsize=1024LL*1024;
bool sortbysize=false;

struct tree_t;
tree_t *mnewtree_t();

struct hash_t {
  unsigned char hash[16];
  bool operator<(const hash_t &h) const {
    for(int i=0;i<16;i++) {
      if(hash[i]<h.hash[i]) return 1;
      if(hash[i]>h.hash[i]) return 0;
    }
    return 0;
  }
  bool operator==(const hash_t &h) const {
    for(int i=0;i<16;i++) {
      if(hash[i]!=h.hash[i]) return 0;
    }
    return 1;
  }
  bool operator!=(const hash_t &h) const {
    for(int i=0;i<16;i++) {
      if(hash[i]!=h.hash[i]) return 1;
    }
    return 0;
  }
  void setx(int x) {
    memset(hash,x,16);
  }
  bool isx(int x) const {
    for(int i=0;i<16;i++) {
      if(hash[i]!=x) return 0;
    }
    return 1;
  }
  bool isx() const {
    if(isx(int(hash[0]))) return 1;
    return 0;
  }
  bool issol() const {
    return isx(0);
  }
  bool isunk() const {
    return isx(1);
  }
  bool isskip() const {
    return isx(2);
  }

  bool setskip() {
    setx(2);
  }
  bool setsol() {
     setx(0);
  }
  bool setunk() {
    setx(1);
  }

  /*
  hash_t(int i=1) {
    setunk();
  }
  hash_t(const hash_t &h) {
    memcpy(hash,h.hash,16);
  }
  */
};

hash_t hx(int i)
{
  hash_t h;
  h.setx(i);
  return h;
}

#define PREAD
//#define ODIRECT

hash_t prehashfile(const string &s,hash_t &h1,ll_t size)
{
  static unsigned char prehash_buff[4096*3];
  static unsigned char *prehash_buff_=(prehash_buff+512)-((reinterpret_cast<ll_t>(prehash_buff+512))&511); // nice !
  
  int fd=open(s.c_str(),O_RDONLY
#ifdef ODIRECT
	      |O_DIRECT
#endif
	      );
  if(fd==-1) return hx(1);

  size_t r;
  unsigned char *buff=prehash_buff_;
  MD5_CTX ctx;
  if(!noend)
    MD5_Init(&ctx);

  MD5_CTX ctx1;
  MD5_Init(&ctx1);

  r=pread(fd,buff,4096,0);
  if(r!=4096) {
    eprintf("error in read %s (prehash 0)\n",s.c_str());
    close(fd);
    h1.setx(1);
    return hx(1);
  }
  if(!noend)
    MD5_Update(&ctx,buff,r);
  MD5_Update(&ctx1,buff,r);

  if(!noend) {
    r=pread64(fd,buff,4096,4096*(size/8192));
    if(r!=4096) {
      eprintf("error in read %s (prehash 1)\n",s.c_str());
      close(fd);
      return hx(1);
    }
    
    MD5_Update(&ctx,buff,r);

    r=pread(fd,buff,4096*2,4096*((size-2048)/4096));
    if(r!=size-4096*((size-2048)/4096)) {
      eprintf("error in read %s (prehash 2)\n",s.c_str());
      close(fd);
      return hx(1);
    }
    
    MD5_Update(&ctx,buff,size-4096*((size-2048)/4096));
  }

  close(fd);
    
  hash_t h;
  MD5_Final(h1.hash,&ctx1);
  if(!noend)
    MD5_Final(h.hash,&ctx);
  else 
    h=h1;
  return h;
}

#if 0
hash_t prehashfile(const string &s,hash_t &h1)
{
  int fd=open(s.c_str(),O_RDONLY);
  if(fd==-1) return hx(1);

  ll_t size=0;
  size_t r;
  if(!noend) {
    size=lseek64(fd,0,SEEK_END);
    if(size<0) {
      eprintf("error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      h1.setx(1);
      return hx(1);
    }
    r=lseek64(fd,0,SEEK_SET);
    if(r<0) {
      eprintf("error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      h1.setx(1);
      return hx(1);
    }
  }
 
  unsigned char buff[4096];
  MD5_CTX ctx;
  if(!noend)
    MD5_Init(&ctx);

  MD5_CTX ctx1;
  MD5_Init(&ctx1);

  r=read(fd,buff,4096);
  if(r!=4096) {
    eprintf("error in read %s (prehash)\n",s.c_str());
    close(fd);
    h1.setx(1);
    return hx(1);
  }
  if(!noend)
    MD5_Update(&ctx,buff,r);
  MD5_Update(&ctx1,buff,r);

  if(!noend) {
    r=lseek64(fd,4096*(size/8192),SEEK_SET);
    if(r<0) {
      eprintf("error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      return hx(1);
    }
    r=read(fd,buff,4096);
    if(r!=4096) {
      eprintf("error in read %s (prehash)\n",s.c_str());
      close(fd);
      return hx(1);
    }
    
    MD5_Update(&ctx,buff,r);
    
    r=lseek64(fd,-4096,SEEK_END);
    if(r<0) {
      eprintf("error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      return hx(1);
    }
    r=read(fd,buff,4096);
    if(r!=4096) {
      eprintf("error in read %s (prehash)\n",s.c_str());
      close(fd);
      return hx(1);
    }
    
    MD5_Update(&ctx,buff,r);
  }

  close(fd);
    
  hash_t h;
  MD5_Final(h1.hash,&ctx1);
  if(!noend)
    MD5_Final(h.hash,&ctx);
  else 
    h=h1;
  return h;
}
#endif

#define HASH_CHK (8*1024*1024)
char hash_buff[HASH_CHK];

hash_t hashfile(const string &s)
{
  int fd=open(s.c_str(),O_RDONLY);
  if(fd==-1) return hx(1);
  
  MD5_CTX ctx;
  MD5_Init(&ctx);

  size_t r;
  do {
    r=read(fd,hash_buff,HASH_CHK);
    if(r==-1) {
      eprintf("error in read %s\n",s.c_str());
      close(fd);
      return hx(1);
    }
    MD5_Update(&ctx,hash_buff,r);
  } while(r==HASH_CHK);

  close(fd);
    
  hash_t h;
  MD5_Final(h.hash,&ctx);
  return h;
}

hash_t hashbuff(const char *bf,int size)
{
  MD5_CTX ctx;
  MD5_Init(&ctx);
  MD5_Update(&ctx,bf,size);
  hash_t h;
  MD5_Final(h.hash,&ctx);
  return h;
}

struct tree_t {
  tree_t *p;
  tree_t *n;
  char *name;
  ll_t size; // -> 4 avec ab aux
  char type; // 0: file 1; dir, 2: other
  unsigned char flags;
  short p_dev; // tab aux
  hash_t hash;

  union {
    struct {
      tree_t *hl; // ->union avec hash
      hash_t hash1; // pas trop utile : tab a cote, virer apres sim_files
      //hash_t prehash; //virer
      ino_t ino;
    } reg;
    
    struct {
      tree_t *e;
      unsigned int lin_i;
      unsigned int lin_s;
    } dir;
  } u;

  static map<dev_t,int> rev_dev;
  static vector<dev_t> vec_dev;
  
  const dev_t &getdev() const {
    assert(isreg());
    return vec_dev[p_dev];
  }
  void setdev(dev_t d) {
    if(rev_dev.find(d)==rev_dev.end()) {
      vec_dev.resize(vec_dev.size()+1);
      vec_dev[vec_dev.size()-1]=d;
      rev_dev[d]=vec_dev.size()-1;
    }
    assert(rev_dev.find(d)!=rev_dev.end());
    assert(rev_dev[d]<32768);
    p_dev=rev_dev[d];
  }
  

#define F_MOD 1
#define F_DEL 2

  bool isdir() const {
    return type==1;
  }

  void initdir()  {
    type=1;
    u.dir.e=0;
  }

  bool isreg() const {
    return type==0;
  }
  
  void setdel() { flags|=F_DEL;setmod();}
  bool isdel() const { return flags&F_DEL;}

  void setmod() { flags|=F_MOD;}
  bool ismod() const { return flags&F_MOD;}
  bool ismod_rec() const {
    if(!isdir()) return ismod();
    if(ismod()) return true;
    tree_t *p=u.dir.e;
    while(p) {
      if(p->ismod()) return true;
      p=p->n;
    }
    return false;
  }

  tree_t() {
    hash.setx(1);
    p=n=NULL;
    size=-1;
    type=-1;
    flags=0;
    //u.reg.prehash.setx(0);
    u.reg.hash1.setx(0);
    u.reg.hl=NULL;
  }

  void changemod() {
    string p=getpath();
    struct stat st;
    int r=lstat(p.c_str(),&st);
    if(S_ISDIR(st.st_mode)) {
      chmod(p.c_str(),(0777&st.st_mode)|S_IWUSR|S_IXUSR);
    } else {
      chmod(p.c_str(),(0777&st.st_mode)|S_IWUSR);
    }
  }

  bool rmrf() {
    if(forcedel) changemod();
    changemod();
    setdel();
    if(isdir()) {
      tree_t *p=u.dir.e;
      while(p) {
	p->rmrf();
	p=p->n;
      }
#if REAL
      if(!fakemod) {
	int r=rmdir(getpath().c_str());
	if(r==0 && mod_log) fprintf(mod_log,"rmdir %s\n",getpath().c_str());
	if(r==-1) fprintf(stderr,"cannot rmdir %s\n",getpath().c_str());
	if(r==-1) return false;
      }
#endif
    } else {
#if REAL
      if(!fakemod) {
	int r=unlink(getpath().c_str());
	if(r==0 && mod_log) fprintf(mod_log,"rm %s\n",getpath().c_str());
	if(r==-1) fprintf(stderr,"cannot unlink %s\n",getpath().c_str());
	if(r==-1) return false;
      }
#endif
    }
    return true;
  }
  
  string getpath() const {
    if(p==NULL) return string(name);
    string r=p->getpath();
    return r+string("/")+string(name);
  }

  hash_t compprehash() {
    //md5 depuis fichier
    string p=getpath();
    hash_t h1;
#ifdef PREAD
    hash_t ph=prehashfile(p,h1,size);
#else
    hash_t ph=prehashfile(p,h1);
#endif
    hash=ph;
    u.reg.hash1=h1;
    return ph;
  }

  bool isparent (tree_t *b) const {
    if(b==NULL) return 0;
    if(this==b) return 1;
    return isparent(b->p);
  }

  hash_t comphash() {
    //md5 depuis fichier
    string p=getpath();
    hash=hashfile(p);
    return hash;
  }

  ll_t getsize() {
    string p=getpath();
    struct stat st;
    int r=lstat(p.c_str(),&st);
    if(r==-1) {
      printf("cannot stat %s\n",p.c_str());
      return -1;
    }
    size=st.st_size;
    return size;
  }

  ll_t compsizedir() {
    if(!isdir()) return size;
    bool ok=1;
    ll_t r=0;
    tree_t *ptr=u.dir.e;
    while(ptr) {
      if(ptr->size<0) ok=0;
      else r+=ptr->size;
      ptr=ptr->n;
    }
    size=r;
    return size;
  }

  hash_t comphashdir() {
    if(!isdir()) return hx(1);

    set<hash_t> sh;
    tree_t *ptr=u.dir.e;
    while(ptr) {
      if(ptr->hash.isunk()) {
	hash.setunk();
	return hash;
      }
      if(ptr->hash.issol()) {
	hash.setsol();
	return hash;
      }
      if(!ptr->hash.isskip())
	sh.insert(ptr->hash);
      ptr=ptr->n;
    }

    if(sh.empty()) {
      hash.setskip();
      return hash;
    }
    
    char *bf=new char[16*sh.size()];
    int k=0;
    set<hash_t>::iterator it;
    for(it=sh.begin();it!=sh.end();++it) {
      memcpy(bf+16*k,it->hash,16);
      k++;
    }
    assert(k==sh.size());
    hash=hashbuff(bf,sh.size()*16);
    delete[]bf;
    return hash;
  }

  ll_t minsize() const {
    return size;
  }
};

#define TP_B (1024*128)

vector<tree_t *> treepool;
int treepool_n=0;

struct ptr_tree_t {
  unsigned int i;
};

tree_t *gettreeptr(int i)
{
  return treepool[i/TP_B]+i%TP_B;
}

void treepool_free() 
{
  for(int i=0;i<treepool_n;i++)
    delete [] treepool[i];
  treepool.resize(0);
}

tree_t *mnewtree_t()
{
  static tree_t *b=NULL;
  static int bs;
  if(b==NULL || bs==0) {
    bs=TP_B;
    b=new tree_t[bs];
    treepool.resize(treepool_n+1);
    treepool[treepool_n++]=b;
  }
  tree_t *r=b;
  b++;
  bs--;
  return r;
}

map<dev_t,int> tree_t::rev_dev;
vector<dev_t> tree_t::vec_dev;


void mds(ll_t &r,tree_t *a,set<hash_t> &s)
{
  if(!a->isdir()) return;
  tree_t *p=a->u.dir.e;
  while(p) {
    if(p->type==0) {
      if(p->hash.issol()) r+=p->size;
      else if(!p->hash.isx()) {
	if(s.find(p->hash)==s.end()) {
	  s.insert(p->hash);
	  r+=p->size;
	}
      }
    }
    if(p->type==1)
      mds(r,p,s);
    p=p->n;
  }
}

ll_t mds(list<tree_t *>la)
{
  ll_t r=0;
  set<hash_t> s;
  list<tree_t *>::iterator it;
  for(it=la.begin();it!=la.end();++it)
    mds(r,*it,s);
  
  return r;
}

void ds(ll_t &r,tree_t *a,set<tree_t*> &s)
{
  if(!a->isdir()) return;
  tree_t *p=a->u.dir.e,*p2;
  while(p) {
    if(p->type==0) {
      p2=p;
      while(p2->u.reg.hl) p2=p2->u.reg.hl;
      if(p2->hash.issol()) r+=p2->size;
      else if(!p2->hash.isx()) {
	if(s.find(p2)==s.end()) {
	  s.insert(p2);
	  r+=p2->size;
	}
      }
    }
    if(p->type==1)
      ds(r,p,s);
    p=p->n;
  }
}

ll_t ds(list<tree_t *>la)
{
  ll_t r=0;
  set<tree_t*> s;
  list<tree_t *>::iterator it;
  for(it=la.begin();it!=la.end();++it)
    ds(r,*it,s);

  return r;
}

ll_t nbrf=0;
ll_t nbrfail=0;
ll_t nbrd=0;
ll_t nbro=0;
tree_t *root;
map<ll_t,list<tree_t*> > mst;
//map<ll_t,list<tree_t*> > mst_d;
map<hash_t,list<tree_t*> > mht;
map<hash_t,list<tree_t*> > mh1t;
map<hash_t,list<tree_t*> > mht_d;

void makemh1t(tree_t *t) {
  if(t->isdir()) {
    tree_t *p=t->u.dir.e;
    while(p) {
      makemh1t(p);
      p=p->n;
    }
    return;
  }
  if(t->type==0 && !t->u.reg.hash1.isx()) 
    mh1t[t->u.reg.hash1].push_back(t);
}

void maketree(const string &path,tree_t *tree) {
  //printf("%s\n",path.c_str());
  if(!tree->isdir()) return;
  
  DIR *d=opendir(path.c_str());
  if(d==NULL) {
    eprintf("cannot opendir %s\n",path.c_str());
    return;
  }
  tree_t *prev=tree;
  struct dirent *e;
  while(1) {
    e=readdir(d);
    if(e==NULL) break;
    if(0==strcmp(e->d_name,".")) continue;
    if(0==strcmp(e->d_name,"..")) continue;
    
    tree_t *ee=mnewtree_t();
    
    ee->p=tree;
    ee->name=mstrdup(e->d_name);
    if(prev==tree) tree->u.dir.e=ee;
    else prev->n=ee;

    struct stat st;
    int r=lstat(ee->getpath().c_str(),&st);
    if(r==-1) {
      eprintf("cannot stat %s\n",ee->getpath().c_str());
      ee->type=-1;
      ee->size=-1;
      nbrfail++;
    } else {
      if(S_ISDIR(st.st_mode)) {
	ee->initdir();
	//ee->type=1;
	nbrd++;
      }
      else if(S_ISREG(st.st_mode)) {
	nbrf++;
	ee->type=0;
	ee->size=st.st_size;
	ee->setdev(st.st_dev);
	ee->u.reg.ino=st.st_ino;

	if(ee->size>0)
	  mst[ee->size].push_back(ee);
	else
	  ee->hash.setskip();
      }
      else {
	nbro++;
	ee->type=2;
	ee->size=0;
	ee->hash.setskip();
      }
    }
    if((nbrfail+nbrf+nbrd+nbro)%10000==0) {
      fprintf(stderr,"%Ld files %Ld dirs %Ld other(s) %Ld fail(s)   \r",nbrf,nbrd,nbro,nbrfail);
    }

    prev=ee;
  }
  closedir(d);

  tree_t *ptr=tree->u.dir.e;
  while(ptr) {
    if(ptr->type==1) {
      maketree(ptr->getpath(),ptr);
    }
    ptr=ptr->n;
  }
  
  ll_t s=tree->compsizedir();
  //printf("sd: %Ld\n",s);
  //if(s>0)   mst_d[s].push_back(tree);
}

tree_t *maketree(const char *path) {
  tree_t *root=mnewtree_t();
  root->initdir();
  root->name=mstrdup(path);
  maketree(string(path),root);
  return root;
}

void aff(tree_t *t) 
{
  if(t->hash.isunk())
    lprintf("%s: %Ld (%s) %d\n",t->getpath().c_str(),t->size,human(t->size),t->hash.isx()?int(t->hash.hash[0]):-1);
  if(!t->isdir()) return;
  tree_t *p=t->u.dir.e;
  while(p) {
    aff(p);
    p=p->n;
  }
}

void fake(tree_t *t) 
{
  if(!t->isdir()) return;
  static ll_t k=0;
  tree_t *p=t->u.dir.e;
  while(p) {
    if(p->isreg()) {
      (*((ll_t*)(p->hash.hash+6)))=1+p->size;
      //(*((ll_t*)(p->u.reg.prehash.hash+6)))=1+p->size;
      (*((ll_t*)(p->u.reg.hash1.hash+6)))=-1-p->size;
    }
    fake(p);
    p=p->n;
  }
}

void fake2(tree_t *t) 
{
  if(!t->isdir()) return;
  static ll_t k=0;
  tree_t *p=t->u.dir.e;
  while(p) {
    if(p->isreg()) {
      (*((ll_t*)(p->hash.hash+6)))=1+k;
      //(*((ll_t*)(p->u.reg.prehash.hash+6)))=1+k;
      (*((ll_t*)(p->u.reg.hash1.hash+6)))=-1-k;
      k++;
    }
    fake2(p);
    p=p->n;
  }
}

void makehashdir(tree_t *t) {
  if(t->type==0 && t->u.reg.hl) {
    tree_t *tt=t->u.reg.hl;
    while(tt->type==0 && tt->u.reg.hl) tt=tt->u.reg.hl;
    t->hash=tt->hash;
    return;
  }
  if(t->type==1) {
    tree_t *p=t->u.dir.e;
    while(p) {
      makehashdir(p);
      p=p->n;
    }
    t->comphashdir();
    if(t->size>0 && !t->hash.isx()) {
      mht_d[t->hash].push_back(t);
    }
  }
} 

multimap<ll_t,tree_t*> mmlof;
void dirwithlotoffiles(tree_t *t) {
  if(t->type==1) {
    ll_t r=0;
    tree_t *p=t->u.dir.e;
    while(p) {
      r++;
      dirwithlotoffiles(p);
      p=p->n;
    }
    if(r>200) mmlof.insert(make_pair(r,t));
  }
} 

bool isexactdir(tree_t *a,tree_t *b) 
{
  if(!a->isdir() || !b->isdir()) return 0;
  map<string,tree_t*> sb;
  tree_t *p,*p2;
  p=b->u.dir.e;
  while(p) {
    sb[p->name]=p;
    p=p->n;
  }
  p=a->u.dir.e;
  while(p) {
    if(sb.find(p->name)==sb.end()) return false;
    p2=sb[p->name];
    if(p->type!=p2->type 
       || (p->type==0 && 
	   (p->size!=p2->size || p->hash!=p2->hash || p->hash.issol()))) return false;
    p=p->n;
  }

  p=a->u.dir.e;
  while(p) {
    p2=sb[p->name];
    if(p->type==1 && isexactdir(p,p2)==false) return false;
    p=p->n;
  }

  return true;
}


/******** SIM DIRS *********/

double cmpdir_factor=0.05;
double cmpdir_factor2=0.65;

namespace sim_dir {
void dir(tree_t *a,map<hash_t,tree_t*> &s, ll_t &r)
{
  if(!a->isdir()) return;
  tree_t *p=a->u.dir.e,*p2;
  while(p) {
    if(p->type==1)
      dir(p,s,r);

    if(p->type==0) {
      p2=p;
      while(p2->u.reg.hl) p2=p2->u.reg.hl;
      if(p2->hash.issol()) r+=p2->size;
      else if(!p2->hash.isx()) {
	if(s.find(p2->hash)==s.end()) {
	  s[p2->hash]=p2;
	}
      }
    }

    p=p->n;
  }
}

void cmpdir(tree_t *a,tree_t *b,ll_t &cap,ll_t &caph,ll_t &amb, ll_t &bma) 
{ 
  if(!a->isdir()) return;
  if(!b->isdir()) return;
  amb=bma=cap=caph=0;
  map<hash_t,tree_t*> sa,sb;
  dir(a,sa,amb);
  dir(b,sb,bma);
  map<hash_t,tree_t*>::iterator it;
  for(it=sa.begin();it!=sa.end();++it) {
    if(sb.find(it->first)==sb.end()) amb+=it->second->size;
    else if(it->second==sb.find(it->first)->second)
      caph+=it->second->size;
    else
      cap+=it->second->size;
  }
  for(it=sb.begin();it!=sb.end();++it)
    if(sa.find(it->first)==sa.end()) bma+=it->second->size;
}

struct cmpdir_t {
  ll_t amb,bma,cap,caph;
  bool disc;
  bool isok(double f) const {
    ll_t i=cap+caph;
    ll_t m=amb;
    if(m<bma) m=bma;
    if(m<f*i) return 1;
    return 0;
  }
};

set<pair<tree_t*,tree_t*> > scmpdir;
map<pair<tree_t*,tree_t*>,cmpdir_t> mcmpdir;

void cmpdirclean() {
  map<pair<tree_t*,tree_t*>,cmpdir_t>::iterator it,j;
  for(it=mcmpdir.begin();it!=mcmpdir.end();++it) {
    tree_t *a=it->first.first;
    tree_t *b=it->first.second;
    for(int i=1;i<4;i++) {
      tree_t *aa=a;
      tree_t *bb=b;
      if(i&1) aa=a->p;
      if(i&2) bb=b->p;
      if(aa==NULL ||bb==NULL) continue;
      if((j=mcmpdir.find(make_pair(aa,bb)))==mcmpdir.end()) continue;
      if(j->second.cap==it->second.cap &&j->second.caph==it->second.caph) j->second.disc=1;
    }
  }
}

void insertcmpdir(tree_t *a,tree_t*b) {
  if(a==NULL || b==NULL || a==b) return;
  if(a->isparent(b) ||b->isparent(a)) return;
  if(!a->isdir()) return;
  if(!b->isdir()) return;
  if(a>b) {insertcmpdir(b,a); return;}
  pair<tree_t*,tree_t*> p=make_pair(a,b);
  if(mcmpdir.find(p)!=mcmpdir.end()) return;
  scmpdir.insert(p);
}

void proccmpdir(tree_t *a,tree_t *b)
{
  double f=cmpdir_factor;
  cmpdir_t c;
  c.amb=c.bma=c.cap=c.caph=0;
  c.disc=0;
  assert(a->type==1 && b->type==1);
  
  if(a->hash!=b->hash || a->hash.isx()) {
    cmpdir(a,b,c.cap,c.caph,c.amb,c.bma);
    mcmpdir[make_pair(a,b)]=c; 
    ll_t i=c.cap+c.caph;
    ll_t m=c.amb;
    if(m>c.bma) m=c.bma;
    if(c.amb<f*i) insertcmpdir(a->p,b);
    if(c.bma<f*i) insertcmpdir(a,b->p);
    if(m<f*i) insertcmpdir(a->p,b->p);
  } else {
    c.cap=-1;
    mcmpdir[make_pair(a,b)]=c;
    insertcmpdir(a,b->p);
    insertcmpdir(a->p,b->p);
    insertcmpdir(a->p,b);
  }
}

void compcmpdir() 
{
  ll_t k=0;
  map<hash_t,list<tree_t*> >::iterator it2;
  
  for(it2=mht.begin();it2!=mht.end();++it2) {
    if(it2->second.size()>1) {
      list<tree_t *>::iterator i,j;
      for(i=it2->second.begin();i!=it2->second.end();++i) {
	for(j=it2->second.begin();j!=i;++j) {
	  insertcmpdir((*i)->p,(*j)->p);
	}
      }
    }
  }

  while(scmpdir.empty()==0) {
    if(k%100==0) fprintf(stderr,"%Ld/%Ld   \r",ll_t(mcmpdir.size()),ll_t(scmpdir.size()));
    k++;
    pair<tree_t*,tree_t*> p=*(scmpdir.begin());
    scmpdir.erase(scmpdir.begin());

    proccmpdir(p.first,p.second);
  }
  fprintf(stderr,"                  \r");
}


void main_search_sim_dirs()
{
  ll_t ti=time(NULL);
  lprintf("search similars dirs...\n");
  compcmpdir();
  cmpdirclean();
  lprintf("%Ld pairs\n",ll_t(mcmpdir.size()));
  lprintf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);
  printmemusage();
 
  /*
    map<pair<tree_t*,tree_t*>,cmpdir_t>::iterator it;
    for(it=mcmpdir.begin();it!=mcmpdir.end();++it) {
    tree_t *a=it->first.first;
    tree_t *b=it->first.second;
    cmpdir_t c=it->second;
    lprintf("%s %s %Ld %Ld %Ld %Ld\n",a->getpath().c_str(),b->getpath().c_str(),c.cap,c.caph,c.amb,c.bma);
    //if(a->hash.isx()) lprintf("a: %d\n",int(a->hash.hash[0]));
    //if(b->hash.isx()) lprintf("b: %d\n",int(b->hash.hash[0]));
    }
  */
  
  map<pair<tree_t*,tree_t*>,cmpdir_t>::iterator it2;
  multimap<ll_t,pair<pair<tree_t*,tree_t*>,cmpdir_t> > mmsh;
  for(it2=mcmpdir.begin();it2!=mcmpdir.end();++it2) {
    if(it2->second.cap==-1) continue;
    if(it2->second.disc) continue;
    if(!it2->second.isok(cmpdir_factor2)) continue;
    ll_t s=it2->second.cap;
    if(!sortbysize) s+=it2->second.caph;
    mmsh.insert(make_pair(s,make_pair(it2->first,it2->second)));
  }
  
  FILE *out=fopen("sim_dirs_old.txt","w");
  multimap<ll_t,pair<pair<tree_t*,tree_t*>,cmpdir_t> >::reverse_iterator itr;
  for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
    if(itr->first<minreportsize) break;
    tree_t *a=itr->second.first.first;
    tree_t *b=itr->second.first.second;
    cmpdir_t c=itr->second.second;
    fprintf(out,"=== save %s ===\n",human(itr->second.second.cap));
    fprintf(out," %Ld (hl: %Ld) < %Ld  > %Ld\n",c.cap,c.caph,c.amb,c.bma);
    fprintf(out,"%s\n%s\n",a->getpath().c_str(),b->getpath().c_str());
  }
  fclose(out);
  
}
}

/******** SIM DIRS 2 *********/

namespace sim_dir_2 {

  struct cmpdir_t {
    ll_t amb,bma,cap,caph;
    bool disc;
    bool isok(double f) const {
      ll_t i=cap+caph;
      ll_t m=amb;
      if(m<bma) m=bma;
      if(m<f*i) return 1;
      return 0;
    }
  };
  
  map<tree_t*,set<tree_t*> > scmpdir;
  map<tree_t*,map<tree_t*,cmpdir_t> > mcmpdir;

  struct lin_t { 
    hash_t h;
    ll_t s;
    tree_t *t;
  };

  lin_t *lin=NULL;
  
  ll_t lin_size(tree_t *ro) 
  {
    if(!ro->isdir()) return 0;
    ll_t r=0;
    ll_t sol=0; 
    tree_t *p=ro->u.dir.e;
    while(p) {
      if(p->type==0) {
	if(p->hash.isx()) {
	  if(p->hash.issol() && p->size>0) sol+=p->size;
	} else 
	  r++;
      }
      if(p->type==1) {
	r+=lin_size(p);
      }
      p=p->n;
    }

    if(sol>0) r++;
    return r;
  }

  ll_t lin_pop(tree_t *ro,lin_t *li) 
  {
    if(!ro->isdir()) return 0;
    ro->u.dir.lin_i=li-lin;
    ll_t r=0;
    ll_t sol=0; 
    tree_t *p=ro->u.dir.e;
    while(p) {
      if(p->type==0) {
	if(p->hash.isx()) {
	  if(p->hash.issol() && p->size>0) sol+=p->size;
	} else {
	  li[r].h=p->hash;
	  li[r].s=p->size;
	  li[r].t=p;
	  r++;
	}
      }
      if(p->type==1) {
	r+=lin_pop(p,li+r);
      }
      p=p->n;
    }
    if(sol>0) {
      li[r].h.setsol();
      li[r].s=sol;
      li[r].t=NULL;
      r++;
    }
    ro->u.dir.lin_s=r;
    return r;
  }
  
  void make_lin(tree_t *root) {
    ll_t k=lin_size(root);
    lin=new lin_t[k];
    lin_pop(root,lin);
  }

  void cmpdir(tree_t *a,tree_t *b,map<hash_t,tree_t*> &sb,ll_t &cap,ll_t &caph,ll_t &amb, ll_t &bma) 
  { 
    amb=bma=cap=caph=0;
    if(!a->isdir()) return;
    if(!b->isdir()) return;

    lin_t *t=lin+a->u.dir.lin_i;
    ll_t s=a->u.dir.lin_s;

    for(int i=0;i<s;i++) {
      if(t[i].t==NULL) amb+=t[i].s;
      else {
	map<hash_t,tree_t*>::iterator it=sb.find(t[i].h);
	if(it==sb.end()) amb+=t[i].s;
	else 
	  if(t[i].t==it->second)
	    caph+=t[i].s;
	  else
	    cap+=t[i].s;
      }
    }
    bma=b->size-cap-caph;
    if(bma<0) bma=0;

    //printf("cmpdir %s %s : %Ld hl: %Ld < %Ld > %Ld\n",a->getpath().c_str(),b->getpath().c_str(),cap,caph,amb,bma);
  }

  void insertcmpdir(tree_t *a,tree_t*b) {
    if(a==NULL || b==NULL || a==b) return;
    if(a->isparent(b) ||b->isparent(a)) return;
    pair<tree_t*,tree_t*> p=make_pair(a,b);
    if(mcmpdir[a].find(b)==mcmpdir[a].end()) scmpdir[a].insert(b);
    if(mcmpdir[b].find(a)==mcmpdir[b].end()) scmpdir[b].insert(a);
  }
  

  void rec(tree_t *r,map<hash_t,tree_t*> &m)
  {
    if(!r->isdir()) return;
    tree_t *p=r->u.dir.e;
    while(p) {
      if(p->type==0) {
	if(p->hash.isx()) {

	} else {
	  m[p->hash]=p;
	}
      }
      if(p->type==1) {
	map<hash_t,tree_t*> m2;
	rec(p,m2);
	map<hash_t,tree_t*>::iterator it;
	for(it=m2.begin();it!=m2.end();++it)
	  m[it->first]=it->second;
      }
      p=p->n;
    }

    if(scmpdir.find(r)==scmpdir.end()) return;
    while(!scmpdir[r].empty()) {
      set<tree_t*>::iterator it=scmpdir[r].begin();
      cmpdir_t c;c.disc=0;
      if(r->hash==(*it)->hash && !r->hash.isx()) {
	c.cap=-1;
      } else {
	if(r->size*3<(*it)->size) {

	} else {
	  cmpdir(*it,r,m,c.cap,c.caph,c.amb,c.bma);
	  mcmpdir[r][*it]=c;

	  //remonte
	  if(c.isok(cmpdir_factor)) {
	    insertcmpdir(r,(*it)->p);
	    insertcmpdir(r->p,(*it));
	    insertcmpdir(r->p,(*it)->p);
	  }
	}
      }
      scmpdir[r].erase(it);
    }
  }
  
  void main_search_sim_dirs()
  {
    lprintf("SIM DIR 2...\n");
    ll_t ti=time(NULL);
    lprintf("lin...\n");
    make_lin(root);
    lprintf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);
    printmemusage();

    {
      ti=time(NULL);
      lprintf("init...\n");

      map<hash_t,list<tree_t*> >::iterator it2;
      multimap<ll_t,hash_t> mmsh;
      multimap<ll_t,hash_t>::reverse_iterator rit;
      ll_t tt=0;

      for(it2=mht.begin();it2!=mht.end();++it2) {
	if(it2->second.size()>1) {
	  ll_t s;
	  s=(*(it2->second.begin()))->size;
	  if(s>1024 && it2->second.size()>15)
	    lprintf("warn: %Ld id_files of size %Ld\n",ll_t(it2->second.size()),s);
	  if(s>1024 && it2->second.size()<=15) {
	    mmsh.insert(make_pair(s,it2->first));
	    tt+=it2->second.size()*(it2->second.size()-1);
	  }
	}
      }
      lprintf("tt=%Ld\n",tt);
      
      for(rit=mmsh.rbegin();rit!=mmsh.rend();++rit) {
	list<tree_t *>::iterator i,j;
	for(i=mht[rit->second].begin();i!=mht[rit->second].end();++i) {
	  for(j=mht[rit->second].begin();j!=i;++j) {
	    insertcmpdir((*i)->p,(*j)->p);
	  }
	}
      }

      tt=0;
      for(it2=mht_d.begin();it2!=mht_d.end();++it2) {
	if(it2->second.size()>1) {
	  ll_t s;
	  s=(*(it2->second.begin()))->size;
	  if(s>1024 && it2->second.size()>15)
	    lprintf("warn: %Ld id_dirs of size %Ld\n",ll_t(it2->second.size()),s);
	  if(s>1024 && it2->second.size()<=15) {
	    mmsh.insert(make_pair(s,it2->first));
	    tt+=it2->second.size()*(it2->second.size()-1);
	  }
	}
      }
      lprintf("tt=%Ld\n",tt);
      
      for(rit=mmsh.rbegin();rit!=mmsh.rend();++rit) {
	list<tree_t *>::iterator i,j;
	for(i=mht_d[rit->second].begin();i!=mht_d[rit->second].end();++i) {
	  for(j=mht_d[rit->second].begin();j!=i;++j) {
	    insertcmpdir((*i)->p,(*j)->p);
	  }
	}
      }
      lprintf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);
      printmemusage();
    }

    lprintf("rec...\n");
    ti=time(NULL);
    map<hash_t,tree_t*> m;
    rec(root,m);
    lprintf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);
    printmemusage();
    
    map<tree_t*,map<tree_t*,cmpdir_t> >::iterator it2;
    map<tree_t*,cmpdir_t>::iterator it3;
    multimap<ll_t,pair<pair<tree_t*,tree_t*>,cmpdir_t> > mmsh;
    for(it2=mcmpdir.begin();it2!=mcmpdir.end();++it2) {
      for(it3=it2->second.begin();it3!=it2->second.end();++it3) {
	if(it3->second.cap==-1) continue;
	if(it3->second.disc) continue;
	if(!it3->second.isok(cmpdir_factor2)) continue;
	ll_t s=it3->second.cap;
	if(!sortbysize) s+=it3->second.caph;
	mmsh.insert(make_pair(s,make_pair(make_pair(it2->first,it3->first),it3->second)));
      }
    }
    
    FILE *out=fopen("sim_dirs.txt","w");
    multimap<ll_t,pair<pair<tree_t*,tree_t*>,cmpdir_t> >::reverse_iterator itr;
    for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
      if(itr->first<minreportsize) break;
      tree_t *a=itr->second.first.first;
      tree_t *b=itr->second.first.second;
      cmpdir_t c=itr->second.second;
      fprintf(out,"=== save %s ===\n",human(itr->second.second.cap));
      fprintf(out," %Ld (hl: %Ld) < %Ld  > %Ld\n",c.cap,c.caph,c.amb,c.bma);
      fprintf(out,"%s\n%s\n",a->getpath().c_str(),b->getpath().c_str());
    }
    fclose(out);

    delete[]lin;
  }
  
}

/**** interactive *******/

#define INT_OK 2
#define INT_CONT 3
#define INT_STOP 1

#define ERR_NOTHING -1
#define ERR_NOENT -2
#define ERR_MOD -3
#define ERR_TYPE -3

#define ERR_PERM -10

int interactive(list<tree_t*> li)
{
  int type=-1;
  dev_t dev;
  bool hl=1;

  //stat files
  map<int,tree_t*> m;
  int k=1;
  list<tree_t*>::iterator it;
  for(it=li.begin();it!=li.end();++it) {
    if((*it)->ismod()) {
      continue;
      return ERR_MOD;
    }
    string fn=(*it)->getpath();
    struct stat st;
    int r=lstat(fn.c_str(),&st);
    if(r==-1) {
      continue;
      return ERR_NOENT;
    }
    if(S_ISREG(st.st_mode) && st.st_size!=(*it)->size) {
      continue;
      //return ERR_MOD;
    }

    if(S_ISREG(st.st_mode)) {
      //printf("%d %Ld\n",int((*it)->p_dev),ll_t((*it)->getdev()));
      if(type==-1) {
	type=0;
	dev=(*it)->getdev();
      }
      else {
	if(type!=0) return ERR_TYPE;
	if(dev!=(*it)->getdev()) hl=0;
      }
    } else if(S_ISDIR(st.st_mode)) {
      hl=0;
      if(type==-1) type=1;
      else if(type!=1) return ERR_TYPE;
    } else 
      return ERR_TYPE;

    m[k++]=*it;
  }
  if(k<=2) return ERR_NOTHING;

  while(1) {
    printf("\n===== SIZE : %s =====\n", human(m[1]->size));
    for(int i=1;i<k;i++) {
      printf("%d) %s",i,m[i]->getpath().c_str());
      if(type==1) {
	if(i>=2) 
	  printf(" %sEXACT",isexactdir(m[i],m[1])?"":"NO-");
      }
      printf("\n");
    }
    printf("\n%s%s",
	   hl?" 'h': hardlink\n":"",
	   " 's x' simlink to x\n"
	   " 'k x': keep only x\n"
	   //	 " 'd x [y ...]': delete x, y...\n"
	   " '!' : stop interactive mode\n"
	   " (empty) : do nothing (skip this entry)\n"
	   "? ");
    fflush(stdout);
    char cmd[100];
    char *rr=fgets(cmd,100,stdin);
    if(rr==NULL) return INT_STOP;
    if(cmd[0]=='!') {
      return INT_STOP;
    }
    if(cmd[0]==0 || cmd[0]=='\n') {
      return INT_OK;
    }
    if(cmd[0]=='h' || cmd[0]=='s' || cmd[0]=='k') {
      int t=1;
      if(cmd[0]=='s' || cmd[0]=='k') {
	int r=sscanf(cmd+1,"%d",&t);
	if(r!=1||t<1||t>=k) continue;
      }
      for(int i=1;i<k;i++) 
	if(i!=t) {
	  m[i]->rmrf();
	  /*
	    int r=unlink(m[i]->getpath().c_str());
	    if(r==0) fprintf(stderr,"cannot unlink %s\n",m[i]->getpath().c_str());
	  */
	}
      if(cmd[0]=='h' || cmd[0]=='s') {
	for(int i=1;i<k;i++) 
	  if(i!=t) {
#if REAL
	    if(!fakemod) {
	      int r;
	      if(cmd[0]=='h') {
		r=link(m[t]->getpath().c_str(),m[i]->getpath().c_str());
		if(r==0) {
		  if(mod_log) fprintf(mod_log,"ln %s %s\n",m[t]->getpath().c_str(),m[i]->getpath().c_str());
		}
	      }
	      if(cmd[0]=='s' || r==-1) {
		if(cmd[0]=='h') fprintf(stderr,"cannot link %s\n",m[i]->getpath().c_str());
		r=symlink(m[t]->getpath().c_str(),m[i]->getpath().c_str());
		if(r==0) {
		  if(mod_log) fprintf(mod_log,"ln -s %s %s\n",m[t]->getpath().c_str(),m[i]->getpath().c_str());
		}
		if(r==-1) {
		  fprintf(stderr,"cannot symlink %s\n",m[i]->getpath().c_str());
		}
	      }
	    }
#endif
	  }
      }
      return INT_OK;
    }
  }
  return 0;
}

/************************/

struct sit_t {
  dev_t dev;
  ino_t ino;
  tree_t *t;
};

int cmpsit(const void *aa,const void *bb)
{
  const sit_t *a=(const sit_t*)aa;
  const sit_t *b=(const sit_t*)bb;

  if(a->dev<b->dev) return -1;
  if(a->dev>b->dev) return 1;

  if(a->ino<b->ino) return -1;
  if(a->ino>b->ino) return 1;

  return 0;
}

int fakehash=0;
bool oldsimdir=false;
bool inter=false;

int main(int ac, char **av)
{
  printf("tree_t: %d\n",int(sizeof(tree_t)));

  if(ac<=1) {
    printusage();
    return 1;
  }

  while(1) {
    if(ac<2) break;
    if(0==strcmp(av[1],"-h")) {printusage(); return 0;}
    if(0==strcmp(av[1],"-H")) {printreadme(); return 0;}
    if(0==strcmp(av[1],"-n")) {md5=0;ac--;av++;continue;}
    if(0==strcmp(av[1],"-t")) {sortbysize=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"-b")) {noend=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"-f")) {cmpdir_factor=atof(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1],"-F")) {cmpdir_factor2=atof(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1],"-r")) {minreportsize=atoll(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1],"-i")) {inter=1;ac-=1;av+=1;continue;}
    if(0==strcmp(av[1],"-s")) {systematic=atoll(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1],"--onlysize")) {fakehash=2;ac--;av++;continue;}
    if(0==strcmp(av[1],"--fakehash")) {fakehash=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"--fake")) {fakemod=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"--oldsimdir")) {oldsimdir=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"--force")) {forcedel=1;ac--;av++;continue;}
    
    break;
  } 

  out_log=fopen("log.txt","w");
  mod_log=fopen("mod.txt","a");

  ll_t ti;
  void *mbrk=sbrk(0);

  {
    root=mnewtree_t();
    root->initdir();
    root->name=mstrdup("");
    tree_t *p=root;
    ll_t sr=0;
    for(int i=1;i<ac;i++) {
      ti=time(NULL);
      lprintf("scan %s\n",av[i]);
      tree_t *q=maketree(av[i]);
      if(p==root) root->u.dir.e=q;
      else p->n=q;
      sr+=q->size;
      p=q;
      lprintf("time: %Ld minute(s)   \n",ll_t(time(NULL)-ti)/60);
      printmemusage();
    }
    root->size=sr;
  }

  lprintf("%Ld files %Ld dirs %Ld other(s) %Ld fail(s)   \n",nbrf,nbrd,nbro,nbrfail);
  lprintf("size: %Ld (%s)\n",root->minsize(),human(root->minsize()));

  {
    lprintf("search hardlinks (hl.txt)...\n");

    map<ll_t,list<tree_t*> >::iterator it;
    multimap<ll_t,pair<tree_t*,tree_t*> > mmhl;
    for(it=mst.begin();it!=mst.end();++it) {
      if(it->first>0 && it->second.size()>1) {
	map<pair<dev_t,ino_t>,list<tree_t*>::iterator > mhl;
	list<tree_t *>::iterator i,i2;
	for(i=it->second.begin();i!=it->second.end();) {
	  assert((*i)->isreg());
	  if(mhl.find(make_pair((*i)->getdev(),(*i)->u.reg.ino))!=mhl.end()) {
	    mmhl.insert(make_pair(it->first,make_pair(*i,*(mhl.find(make_pair((*i)->getdev(),(*i)->u.reg.ino))->second))));
	    
	    (*i)->u.reg.hl=*mhl.find(make_pair((*i)->getdev(),(*i)->u.reg.ino))->second;
	    
	    i2=i;++i2;
	    it->second.erase(i);
	    i=i2;
	  } else {
	    mhl[make_pair((*i)->getdev(),(*i)->u.reg.ino)]=i;
	    ++i;
	  }
	}
      }
    }

    FILE *out=fopen("hl.txt","w");
    multimap<ll_t,pair<tree_t*,tree_t*> >::reverse_iterator itr;
    for(itr=mmhl.rbegin();itr!=mmhl.rend();++itr) {
      if(itr->first<minreportsize) break;
      fprintf(out,"  %s\n= %s\n",itr->second.first->getpath().c_str(),itr->second.second->getpath().c_str());
    }
    fclose(out);
  }

  ll_t bth=0;
  ll_t fth=0;

  {
    map<ll_t,list<tree_t*> >::iterator it;
    for(it=mst.begin();it!=mst.end();++it) {
      if((it->first>0 && it->second.size()>1)
	 || it->first>=systematic ) {
	fth+=it->second.size();
	bth+=it->second.size()*it->first;
      }
    }
    lprintf("to prehash: %Ld files\n",fth);
  }

  ll_t bth_c=0;
  ll_t fth_c=0;
  int plog=1;
  ll_t ll=0;
    
  if(fakehash) {
    md5=0;
    if(fakehash==1)
      fake2(root);
    else
      fake(root);
  } else {

  lprintf("sort files by dev/ino...\n");

  sit_t *sit=new sit_t[fth];
  ll_t k=0;

  map<ll_t,list<tree_t*> >::reverse_iterator rit;
  for(rit=mst.rbegin();rit!=mst.rend();++rit) {
    map<hash_t,list<tree_t*> >::iterator pit;
    list<tree_t *>::iterator i;
    if((rit->first>0 && rit->second.size()>1) 
       || rit->first>=systematic ) {
      for(i=rit->second.begin();i!=rit->second.end();++i) {
	assert((*i)->isreg());
	sit[k].dev=(*i)->getdev();
	sit[k].ino=(*i)->u.reg.ino;
	sit[k].t=(*i);
	k++;
      }
    } else  {
      for(i=rit->second.begin();i!=rit->second.end();++i) {
	assert((*i)->isreg());
	if(rit->first==0) {
	  (*i)->hash.setskip();
	  (*i)->u.reg.hash1.setskip();
	  //(*i)->u.reg.prehash.setskip();
	} 
	else {
	  (*i)->hash.setsol();
	  (*i)->u.reg.hash1.setsol();
	  //(*i)->u.reg.prehash.setsol();
	}
      }
    }
  }

  assert(k==fth);
  lprintf("qsort...\n");
  qsort(sit,fth,sizeof(sit_t),cmpsit);
  
  lprintf("prehash...\n");
  ti=time(NULL);
  for(ll_t l=0;l<fth;l++) {
    if(l%1000==0) fprintf(stderr,"%Ld/%Ld  \r",l,fth);
    tree_t *t=sit[l].t;
    if(t->size<4096*5) {
      t->comphash();
      //t->u.reg.prehash=t->hash;
    } else {
      t->compprehash();
    }
  }
  lprintf("time: %Ld minutes   \n",ll_t(time(NULL)-ti)/60);
  printmemusage();
  delete[]sit;
  }

  lprintf("hash...   \n");
  ti=time(NULL);
  map<ll_t,list<tree_t*> >::reverse_iterator rit;
  for(rit=mst.rbegin();rit!=mst.rend();++rit) {
    map<hash_t,list<tree_t*> >::iterator pit;
    list<tree_t *>::iterator i;
    if(rit->first<10*1024*1024) {
      plog=0;
      if(ll<bth_c-10*1024*1024) {
	plog=0;
	char bf2[30];
	fprintf(stderr,"%Ld (%s) : %Ld files %Ld bytes (%s)  \r",rit->first,human(rit->first),fth_c,ll=bth_c,human(bth_c,bf2));
      }
    }

    if(rit->first>0 && rit->second.size()>1) {
      if(rit->first>=4096*5) {
	map<hash_t,list<tree_t*> > mpht;
	for(i=rit->second.begin();i!=rit->second.end();++i) {
	  assert((*i)->isreg());
	  hash_t ph=(*i)->hash;
	  mpht[ph].push_back(*i);
	}
	for(pit=mpht.begin();pit!=mpht.end();++pit) {
	  if(pit->second.size()>1) {
	    for(i=pit->second.begin();i!=pit->second.end();++i) {
	      assert((*i)->isreg());
	      if(md5) 
		(*i)->comphash();
	      else {
		// computes a 'fake hash' with prehash and size
		//(*i)->hash=(*i)->u.reg.prehash;
		*(ll_t*)((*i)->hash.hash)=(*(ll_t*)((*i)->hash.hash))^((*i)->size);
	      }

	      fth_c+=1;
	      bth_c+=rit->first;
	      if(plog) {
		char bf2[10];
		fprintf(stderr,"%Ld (%s) : %Ld files %Ld bytes (%s)  \r",rit->first,human(rit->first),fth_c,bth_c,human(bth_c,bf2));
	      }

	      if(!(*i)->hash.isx()) {
		mht[(*i)->hash].push_back(*i);
	      }
	    }
	  } else {
	    for(i=pit->second.begin();i!=pit->second.end();++i) {
	      (*i)->hash.setsol();
	    }
	    fth_c+=1;
	    bth_c+=rit->first;
	    if(plog) {
	      char bf2[10];
	      fprintf(stderr,"%Ld (%s) : %Ld files %Ld bytes (%s)  \r",rit->first,human(rit->first),fth_c,bth_c,human(bth_c,bf2));
	    }
	  }
	} 
     } else {
	for(i=rit->second.begin();i!=rit->second.end();++i) {

	  fth_c+=1;
	  bth_c+=rit->first;

	  if(!(*i)->hash.isx()) {
	    mht[(*i)->hash].push_back(*i);
	  }
	}
      }
    } else  {
      for(i=rit->second.begin();i!=rit->second.end();++i) {
	if(rit->first==0)
	  (*i)->hash.setskip();
	else
	  (*i)->hash.setsol();
      }
    }
  } 
  mst.clear();
  lprintf("time: %Ld minutes       \n",ll_t(time(NULL)-ti)/60);
  printmemusage();

  multimap<ll_t,hash_t> mmsh_files;

  {
    map<hash_t,list<tree_t*> >::iterator it2;
    
    ll_t rec=0;
    for(it2=mht.begin();it2!=mht.end();++it2) {
      if(it2->second.size()>1) {
	rec+=(it2->second.size()-1)*(*it2->second.begin())->size;
      }
    }
    lprintf("max recuperable : %s\n", human(rec));
    
    lprintf("write id_files.txt\n");

    for(it2=mht.begin();it2!=mht.end();++it2) {
      if(it2->second.size()>1) {
	ll_t s;
	s=(*(it2->second.begin()))->size;
	mmsh_files.insert(make_pair(s,it2->first));
      }
    }

    /*
    {
      multimap<ll_t,hash_t>::reverse_iterator itr;
      for(itr=mmsh_files.rbegin();itr!=mmsh_files.rend();++itr) {
	if(itr->first<minreportsize) break;
	list<tree_t*>::iterator i;
	for(i=mht[itr->second].begin();i!=mht[itr->second].end();++i) {
	}
      }
    }
    */
    
    FILE *out=fopen("id_files.txt","w");
    multimap<ll_t,hash_t>::reverse_iterator itr;
    for(itr=mmsh_files.rbegin();itr!=mmsh_files.rend();++itr) {
      if(itr->first<minreportsize) break;
      list<tree_t*>::iterator i;
      for(i=mht[itr->second].begin();i!=mht[itr->second].end();++i) {
	fprintf(out,"%s ( %Ld = %s )\n",(*i)->getpath().c_str(),(*i)->size,human((*i)->size));
      }
      fprintf(out,"====================\n");
    }
    fclose(out);

  }

  {
    makemh1t(root);
    map<hash_t,list<tree_t*> >::iterator it2;
    
    lprintf("write sim_files.txt %Ld\n",ll_t(mh1t.size()));
    multimap<ll_t,map<hash_t,list<tree_t*> > > mmsh;

    for(it2=mh1t.begin();it2!=mh1t.end();++it2) {
      if(it2->second.size()>1) {
	list<tree_t *>::iterator i;
	map<hash_t,list<tree_t*> > ml;
	ll_t s=-1,s2=-1;
	int k=1;
	for(i=it2->second.begin();i!=it2->second.end();++i) {
	  if((*i)->hash.issol()) {
	    hash_t h=(*i)->hash;
	    *(int*)h.hash=(*(int*)h.hash)^k; // fake hash for 'sol' files
	    k++;
	    ml[h].push_back(*i);
	  } else {
	    if(!(*i)->hash.isx())
	      ml[(*i)->hash].push_back(*i);
	  }
	  if(s==-1 || (*i)->size>s) {s2=s;s=(*i)->size;}
	  else
	    if(s2==-1 || (*i)->size>s2) {s2=(*i)->size;}
	}
	if(ml.size()>1) {
	  assert(s2>0);
	  mmsh.insert(make_pair(s2,ml));
	}
      }
    }
    
    FILE *out=fopen("sim_files.txt","w");
    multimap<ll_t,map<hash_t,list<tree_t*> > >::reverse_iterator itr;
    for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
      if(itr->first<minreportsize) break;
      map<hash_t,list<tree_t*> >::iterator i;
      fprintf(out,"=== size = %Ld = %s ===\n",itr->first,human(itr->first));
      for(i=itr->second.begin();i!=itr->second.end();++i) {
	if(i!=itr->second.begin()) 
	  fprintf(out,"------\n");
	list<tree_t*>::iterator j;
	for(j=i->second.begin();j!=i->second.end();++j) {
	  fprintf(out,"%s ( %Ld = %s )\n",(*j)->getpath().c_str(),(*j)->size,human((*j)->size));
	}
      }
      fprintf(out,"====================\n");
    }
    fclose(out);
    mh1t.clear();
  }

  ti=time(NULL);
  lprintf("hash: dirs...\n");
  makehashdir(root);
  lprintf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);
  printmemusage();

  aff(root);

  //lprintf("%s (%Ld)\n",(*i)->getpath().c_str(),(*i)->size);

  multimap<ll_t,hash_t> mmsh_dirs;

  lprintf("write id_dirs.txt\n");
  {
    map<hash_t,list<tree_t*> >::iterator it2;

    for(it2=mht_d.begin();it2!=mht_d.end();++it2) {
      if(it2->second.size()>1) {
	ll_t s;
	s=(*(it2->second.begin()))->size;
	if(!sortbysize) s=ds(it2->second)-mds(it2->second);
	mmsh_dirs.insert(make_pair(s,it2->first));
      }
    }

    FILE *out=fopen("id_dirs.txt","w");
    multimap<ll_t,hash_t>::reverse_iterator itr;
    for(itr=mmsh_dirs.rbegin();itr!=mmsh_dirs.rend();++itr) {
      if(itr->first<minreportsize) break;
      fprintf(out,"=== save %s ===\n",human(ds(mht_d[itr->second])-mds(mht_d[itr->second])));
      list<tree_t*>::iterator i;
      for(i=mht_d[itr->second].begin();i!=mht_d[itr->second].end();++i) {
	fprintf(out,"%s ( %s%s %Ld = %s )\n",(*i)->getpath().c_str(),i==mht_d[itr->second].begin()||isexactdir(*(mht_d[itr->second].begin()),*i)?"":"NO-",i==mht_d[itr->second].begin()?"":"EXACT",(*i)->size,human((*i)->size));
      }
    }
    fclose(out);
  }

  {
    dirwithlotoffiles(root);
    FILE *out=fopen("lof_dirs.txt","w");
    lprintf("search for dirs with a lot of files (lof_dirs.txt)\n");
    multimap<ll_t,tree_t*>::reverse_iterator itr;
    for(itr=mmlof.rbegin();itr!=mmlof.rend();++itr) {
      fprintf(out,"%s ( %Ld files, %s )\n",itr->second->getpath().c_str(),itr->first,human(itr->second->size));
    }
    fclose(out);
    mmlof.clear();
  }

  if(inter) {
    if(mod_log) fprintf(mod_log,"### Start interactive mode ###\n");

    multimap<ll_t,hash_t>::reverse_iterator itr_f=mmsh_files.rbegin();
    multimap<ll_t,hash_t>::reverse_iterator itr_d=mmsh_dirs.rbegin();
    while(itr_f!=mmsh_files.rend() || itr_d!=mmsh_dirs.rend()) {
      int r;
      if(itr_f==mmsh_files.rend() || 
	 (itr_d!=mmsh_dirs.rend() && itr_d->first>itr_f->first)) {
	do {
	  r=interactive(mht_d[itr_d->second]);
	} while (r==INT_CONT);
	if(r==INT_STOP) break;
	itr_d++;
      } else {
	do {
	  r=interactive(mht[itr_f->second]);
	} while (r==INT_CONT);
	if(r==INT_STOP) break;
	itr_f++;
      }
    }
  }
  
  sim_dir_2::main_search_sim_dirs();

  if(oldsimdir)
    sim_dir::main_search_sim_dirs();
		
  if(out_log)
    fclose(out_log);

  if(mod_log)
    fclose(mod_log);

  treepool_free();
  mstrdup_free();

  return 0;
}
